#!/usr/bin/env scheme-script
;; Copyright © 2019 Göran Weinholt        !#
;; SPDX-License-Identifier: MIT

;; Tests the MBR library. Test images are built by run-test.sh.

(import
  (rnrs)
  (fs partitions common)
  (fs partitions mbr))

(define (make-sector-reader p sector-size)
  (lambda (lba)
    (set-port-position! p (* lba sector-size))
    (get-bytevector-n p sector-size)))

(display "Checking lba->chs\n")
(assert (chs=? (lba->chs 0) (make-chs 0 0 1)))
(assert (chs=? (lba->chs 1) (make-chs 0 0 2)))
(assert (chs=? (lba->chs 30) (make-chs 0 0 31)))
(assert (chs=? (lba->chs 62) (make-chs 0 0 63)))
(assert (chs=? (lba->chs 63) (make-chs 0 1 1)))
(assert (chs=? (lba->chs 64) (make-chs 0 1 2)))
(assert (chs=? (lba->chs 100) (make-chs 0 1 38)))
(assert (chs=? (lba->chs 128) (make-chs 0 2 3)))
(assert (chs=? (lba->chs 200) (make-chs 0 3 12)))
(assert (chs=? (lba->chs 16328) (make-chs 1 4 12)))
(newline)

(display "Checking sexp->mbr disk-id\n")
(let ((mbr (sexp->mbr '((disk-id #x11223344)))))
  (validate-parttable mbr)
  (assert (eqv? #x11223344 (parttable-disk-id mbr))))

(display "Checking sexp->mbr empty list\n")
(let ((mbr (sexp->mbr '())))
  (assert (eqv? 0 (parttable-disk-id mbr))))

(display "Checking sexp->mbr with 1 partition\n")
(let ((mbr (sexp->mbr '((disk-id #x11223344) (part (start 1) (size 10) (type 2))))))
  (validate-parttable mbr)
  (assert (eqv? #x11223344 (parttable-disk-id mbr)))
  (let ((part (car (parttable-partitions mbr))))
    (assert (not (mbrpart-bootable? part)))
    (assert (eqv? 1 (part-number part)))
    (assert (eqv? 1 (part-start-lba part)))
    (assert (eqv? 10 (part-end-lba part)))
    (assert (eqv? 2 (part-type part)))))

(display "Checking sexp->mbr with 1 bootable partition\n")
(let ((mbr (sexp->mbr '((disk-id #x11223344) (part (boot? #t) (start 1) (size 10) (type 2))))))
  (validate-parttable mbr)
  (assert (eqv? #x11223344 (parttable-disk-id mbr)))
  (let ((part (car (parttable-partitions mbr))))
    (assert (mbrpart-bootable? part))
    (assert (eqv? 1 (part-number part)))
    (assert (eqv? 1 (part-start-lba part)))
    (assert (eqv? 10 (part-end-lba part)))
    (assert (eqv? 2 (part-type part)))))

(display "Checking sexp->mbr with 1 partition type name\n")
(let ((mbr (sexp->mbr '((disk-id #x11223344) (part (start 1) (size 10) (type linux))))))
  (validate-parttable mbr)
  (assert (eqv? #x11223344 (parttable-disk-id mbr)))
  (let ((part (car (parttable-partitions mbr))))
    (assert (eqv? 1 (part-number part)))
    (assert (eqv? 1 (part-start-lba part)))
    (assert (eqv? 10 (part-end-lba part)))
    (assert (eqv? #x83 (part-type part)))))

(display "Checking sexp->mbr with default type\n")
(let ((mbr (sexp->mbr '((part (start 1) (size 10))))))
  (validate-parttable mbr)
  (assert (eqv? 0 (parttable-disk-id mbr)))
  (let ((part (car (parttable-partitions mbr))))
    (assert (eqv? 1 (part-number part)))
    (assert (eqv? 1 (part-start-lba part)))
    (assert (eqv? 10 (part-end-lba part)))
    (assert (eqv? 0 (part-type part)))))

(display "Checking sexp->mbr with 2 partitions\n")
(let ((mbr (sexp->mbr '((disk-id #x11223344)
                         (part (start 1) (size 10) (type 2))
                         (part (start 11) (size 20))))))
  (validate-parttable mbr)
  (assert (eqv? #x11223344 (parttable-disk-id mbr)))
  (let ((part (car (parttable-partitions mbr))))
    (assert (eqv? 1 (part-number part)))
    (assert (eqv? 1 (part-start-lba part)))
    (assert (eqv? 10 (part-end-lba part)))
    (assert (eqv? 2 (part-type part))))
  (let ((part (cadr (parttable-partitions mbr))))
    (assert (eqv? 2 (part-number part)))
    (assert (eqv? 11 (part-start-lba part)))
    (assert (eqv? 30 (part-end-lba part)))
    (assert (eqv? 0 (part-type part)))))
(newline)

(display "Checking sexp->mbr with 2 partitions and end sectors\n")
(let ((mbr (sexp->mbr '((disk-id #x11223344)
                         (part (start 1)  (end 32) (type 17))
                         (part (start 33) (end 2880) (type 5))))))
  (validate-parttable mbr)
  (assert (eqv? #x11223344 (parttable-disk-id mbr)))
  (let ((part (car (parttable-partitions mbr))))
    (assert (eqv? 1 (part-number part)))
    (assert (eqv? 1 (part-start-lba part)))
    (assert (eqv? 32 (part-end-lba part)))
    (assert (eqv? 32 (part-total-lba part)))
    (assert (eqv? 17 (part-type part))))
  (let ((part (cadr (parttable-partitions mbr))))
    (assert (eqv? 2 (part-number part)))
    (assert (eqv? 33 (part-start-lba part)))
    (assert (eqv? 2880 (part-end-lba part)))
    (assert (eqv? 2848 (part-total-lba part)))
    (assert (eqv? 5 (part-type part)))))
(newline)

(display "Checking sexp->mbr with 2 partitions and garbage\n")
(let ((mbr (sexp->mbr '((disk-id #x11223344) some-prop
                         (some-other-list a b c)
                         (part (start 1) (size 10) (type 2))
                         (part (start 11) (size 20))))))
  (validate-parttable mbr)
  (assert (eqv? #x11223344 (parttable-disk-id mbr)))
  (let ((part (car (parttable-partitions mbr))))
    (assert (eqv? 1 (part-number part)))
    (assert (eqv? 1 (part-start-lba part)))
    (assert (eqv? 10 (part-end-lba part)))
    (assert (eqv? 2 (part-type part))))
  (let ((part (cadr (parttable-partitions mbr))))
    (assert (eqv? 2 (part-number part)))
    (assert (eqv? 11 (part-start-lba part)))
    (assert (eqv? 30 (part-end-lba part)))
    (assert (eqv? 0 (part-type part)))))
(newline)

(define (make-bytevector-writer disk)
  (lambda (data sector)
    (bytevector-copy! data 0 disk (fx* sector 512) (bytevector-length data))))

(define (make-bytevector-reader disk)
  (let ((tmp (make-bytevector 512 0)))
    (lambda (sector)
      (bytevector-copy! disk (fx* sector 512) tmp 0 512)
      tmp)))

(display "Checking write simple mbr table\n")
(let ((disk (make-bytevector 512 0))
      (mbr (sexp->mbr '((disk-id #x11223344)
                         (part (start 10)  (size 10) (type #x83))
                         (part (start 21) (size 20) (type #x82))))))
  (validate-parttable mbr)
  (create-mbr! mbr (make-bytevector-reader disk) (make-bytevector-writer disk))
  (let ((mbr (read-mbr (make-bytevector-reader disk))))
    (assert (eqv? (parttable-disk-id mbr) #x11223344))
    (assert (eqv? 2 (length (parttable-partitions mbr))))
    (let ((part (car (parttable-partitions mbr))))
      (display (part-end-lba part))(newline)
      (assert (eqv? 1 (part-number part)))
      (assert (eqv? 10 (part-start-lba part)))
      (assert (eqv? 19 (part-end-lba part)))
      (assert (eqv? 10 (part-total-lba part)))
      (assert (eqv? #x83 (part-type part))))
    (let ((part (cadr (parttable-partitions mbr))))
      (assert (eqv? 2 (part-number part)))
      (assert (eqv? 21 (part-start-lba part)))
      (assert (eqv? 40 (part-end-lba part)))
      (assert (eqv? 20 (part-total-lba part)))
      (assert (eqv? #x82 (part-type part))))))
(newline)

(display "Checking testdisk1.img\n")
(call-with-port (open-file-input-port "testdisk1.img")
  (lambda (p)
    (let ((mbr (read-mbr (make-sector-reader p 512))))
      (write mbr)
      (newline)
      (assert (eqv? 1 (length (parttable-partitions mbr))))
      (let ((part (car (parttable-partitions mbr))))
        (assert (eqv? 1 (part-number part)))
        (assert (eqv? 1 (part-start-lba part)))
        (assert (eqv? 2880 (part-end-lba part)))
        (assert (eqv? 2880 (part-total-lba part)))
        (assert (eqv? 1 (part-type part)))))))
(newline)

(define (part->list part)
  (list 'number (part-number part)
        'start (part-start-lba part)
        'end (part-end-lba part)
        'total (part-total-lba part)
        'id (part-type part)))

;; One logical partition
(display "Checking testdisk2.img\n")
(call-with-port (open-file-input-port "testdisk2.img")
  (lambda (p)
    (let ((mbr (read-mbr (make-sector-reader p 512))))
      (write mbr)
      (newline)
      (write (map part->list (parttable-partitions mbr)))
      (newline)
      (assert (equal? (map part->list (parttable-partitions mbr))
                      '((number 1 start 1 end 32 total 32 id 17)
                        (number 2 start 33 end 2880 total 2848 id 5)
                        (number 5 start 34 end 2880 total 2847 id 1)))))))
(newline)

;; Many logical partitions
(display "Checking testdisk3.img\n")
(call-with-port (open-file-input-port "testdisk3.img")
  (lambda (p)
    (let ((mbr (read-mbr (make-sector-reader p 512))))
      (write mbr)
      (newline)
      (write (map part->list (parttable-partitions mbr)))
      (newline)
      (assert
       (equal? (map part->list (parttable-partitions mbr))
               '((number 1 start 40 end 2880 total 2841 id 5)
                 (number 5 start 41 end 100 total 60 id #x83)
                 (number -6 start 101 end 200 total 100 id 5)
                 (number 6 start 102 end 200 total 99 id #x83)
                 (number -7 start 201 end 300 total 100 id 5)
                 (number 7 start 202 end 300 total 99 id #x83)
                 (number -8 start 301 end 1302 total 1002 id 5)
                 (number 8 start 302 end 1302 total 1001 id #x82)
                 (number -9 start 1303 end 2880 total 1578 id 5)
                 (number 9 start 1304 end 2880 total 1577 id #x83)
                 (number 2 start 2 end 30 total 29 id #x83)))))))
(newline)

;; Test with a large sector size
(display "Checking testdisk4.img\n")
(call-with-port (open-file-input-port "testdisk4.img")
  (lambda (p)
    (let ((mbr (read-mbr (make-sector-reader p 4096))))
      (write mbr)
      (newline)
      (write (map part->list (parttable-partitions mbr)))
      (newline)
      (assert
       (equal? (map part->list (parttable-partitions mbr))
               '((number 1 start 101 end 359 total 259 id 5)
                 (number 5 start 110 end 220 total 111 id 131)
                 (number -6 start 102 end 359 total 258 id 5)
                 (number 6 start 103 end 359 total 257 id 130)
                 (number 4 start 1 end 100 total 100 id 1)))))))
(newline)

;; Test with a protective MBR
(display "Checking testdisk1g.img\n")
(call-with-port (open-file-input-port "testdisk1g.img")
  (lambda (p)
    (let ((mbr (read-mbr (make-sector-reader p 512))))
      (write mbr)
      (newline)
      (write (map part->list (parttable-partitions mbr)))
      (newline)
      (assert
       (equal? (map part->list (parttable-partitions mbr))
               '((number 1 start 1 end 2880 total 2880 id #xEE)))))))
(newline)

(flush-output-port (current-output-port))
